/**************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto , (c)2004 by 421                 *
*   ikemo@wakaba.jp                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "datmanager.h"
#include "libkita/qcp932codec.h"
#include "libkita/comment.h"
#include "libkita/access.h"
#include "kitaconfig.h"
#include "kita-utf8.h"
#include <qregexp.h>
#include <qsemaphore.h>

using namespace Kita;

/* UTF-16 */
#define UTF16_BRACKET  0xFF1E /* > */
#define UTF16_0 0xFF10 /* 0 */
#define UTF16_9 0xFF19 /* 9 */
#define UTF16_EQ 0xFF1D /* = */
#define UTF16_COMMA 0xFF0C /* , */

#define KITA_RESDIGIT 4

/*---------------------------------------------------*/

/* DatInfo stores information of *.dat */

DatInfo::DatInfo( const KURL& url ) : m_url( url )
{

    QString idstr;
    KURL logurl = url;
    logurl.setProtocol( "k2ch" );
    QString rawdata = Kita::Access::getCacheData( logurl );
    m_linedat.clear();
    m_subject = QString::null;
    m_maxNum = 0;
    m_spacestr = QString::null;
    m_anclistmap.clear();    

    if ( rawdata.length() ) {
        QCp932Codec codec;
        QString linedata = codec.toUnicode( rawdata );

        QStringList lines = QStringList::split( "\n", linedata );
        for ( QStringList::iterator it = lines.begin(); it != lines.end(); ++it, ++m_maxNum ) {

            /* copy dat */
            m_linedat += ( *it );

            /* get id & copy */
            idstr = "none";
            QStringList list = QStringList::split( "<>", ( *it ), true );
            if ( list.size() == 5 ) {
                /* copied from Comment::parseDateId */
                QRegExp regexp( "(\\d\\d)/(\\d\\d)/(\\d\\d) (\\d\\d):(\\d\\d)( ID:(.*))?" );
                if ( regexp.search( list[ 2 ] ) != -1 ) idstr = regexp.cap( 7 );

                /* get subject */
                if ( m_subject == QString::null && list[ 4 ] != QString::null ) m_subject = list[ 4 ];
            }
            m_idlist += idstr;
        }
    }
}


const QString& DatInfo::getDat( int num )
{

    if ( num > m_maxNum ) return QString::null;

    if ( m_linedat.at( num - 1 ) == m_linedat.end() ) return QString::null;

    return *( m_linedat.at( num - 1 ) );
}


const QString& DatInfo::getId( int num )
{

    if ( num > m_maxNum ) return QString::null;

    if ( m_idlist.at( num - 1 ) == m_linedat.end() ) return QString::null;

    return *( m_idlist.at( num - 1 ) );
}

/* convert dat to HTML */
QString DatInfo::getHtml(int num){

    QString retstr = QString::null;
    QString line = getDat(num);
    bool showAddr = KitaConfig::showMailAddress();
    
    if(line != QString::null){

	Kita::Comment comment( line );

	/* check abone */
	if(DatManager::checkAbone(getId(num),ABONECHK_BY_ID)) return retstr;
	if(DatManager::checkAbone(comment.getName(),ABONECHK_BY_NAME)) return retstr;
	if(DatManager::checkAbone(comment.getBody(),ABONECHK_BY_WORD)) return retstr;

	retstr = comment.toHtml( num, showAddr );
    }

    return retstr;
}


/* Check if No.num has anchors to No.target */
/* For exsample, if No.num has an anchor ">>target",
   then return TRUE.                               */
bool DatInfo::checkRes(const int target, const int num )
{
    AncList anclist;
    AncListMap::Iterator ancmapit = m_anclistmap.find(num);

    /* If map is not set, then do parsing */
    if(ancmapit == m_anclistmap.end() ){

	QString line = getDat(num);
	if(line == QString::null) return FALSE;

	ANCNUM anctmp;
	QString linkstr;
	int refNum[2];
	unsigned int pos;

	/* dummy */
	anclist.clear();
	anctmp.from = 0;
	anctmp.to = 0;
	anclist += anctmp;

	/* remove HTML tags */
	QRegExp rex( "<[^>]*>" );
	line.remove(rex);

	const QChar *chpt = line.unicode();
	unsigned int i;
	unsigned int length = line.length();

	for ( i = 0 ; i < length ; i++ ) {

	    if( chpt[ i ].unicode() == UTF16_BRACKET ||
		(chpt[ i] == '&' && chpt[ i+1] == 'g' && chpt[ i+2] == 't' && chpt[ i+3] == ';')){

		while(DatManager::parseResAnchor(chpt+i,length-i,linkstr,refNum,pos)){
		    if(refNum[1] < refNum[0]) refNum[1] = refNum[0];
		    anctmp.from = refNum[0];
		    anctmp.to = refNum[1];
		    anclist += anctmp;
		    i += pos;
		}

		i += (pos-1);
	    }
	}

	ancmapit = m_anclistmap.insert(num,anclist);
    }
    /* end parsing   */
    
    for ( AncList::iterator it = (*ancmapit).begin(); it != (*ancmapit).end(); ++it ){
	if ( target >= (*it).from && target <= (*it).to ) return TRUE;
    }
    
    return FALSE;
}


/* Get HTML document of res tree.*/
/* For example,

>>1
|-->>4
|  |--->>10
|
|-->>20

*/
QString DatInfo::getTreeByRes(const int rootnum, int& num, QString prestr )
{
    QString retstr = QString::null ;    
    num = 0;
    int maxNum = getMaxResNumber();
    QStringList strlists;

    if(m_spacestr == QString::null){
	QTextCodec * codec = QTextCodec::codecForName( "utf8" );
	m_spacestr = ". ";
	m_framestr1 = codec->toUnicode( KITAUTF8_FRAME1 ); /* |  */
	m_framestr2 = codec->toUnicode( KITAUTF8_FRAME2 ); /* |- */
	m_framestr3 = codec->toUnicode( KITAUTF8_FRAME3 ); /* L  */
    }

    /* find responses that refer to rootnum */
    for ( int i = rootnum+1; i <= maxNum; i++ ){
	if(checkRes(rootnum,i)){
	    num ++;
	    strlists += QString().setNum( i );
	}
    }

    /* make HTML document */
    if(num){
	
	for ( QStringList::iterator it = strlists.begin(); it != strlists.end(); ++it ){
	    QString tmpstr;
	    if((*it) == strlists.last()) tmpstr = m_framestr3;
	    else tmpstr = m_framestr2;
	    retstr += prestr + tmpstr + "<a href=\"#" + (*it) + "\">&gt;&gt;" + (*it) + "</a><br>";

	    /* call myself recursively */
	    if((*it) == strlists.last()) tmpstr = prestr + m_spacestr + m_spacestr + m_spacestr;
	    else tmpstr = prestr + m_framestr1  + m_spacestr;
	    int tmpnum;
	    retstr += getTreeByRes((*it).toInt(),tmpnum,tmpstr);
	    num += tmpnum;
	}
    }

    return retstr;
}


/*---------------------------------------------------*/

/* DatManager manages all information about *.dat. */

DatInfoList DatManager::m_datInfo;
QSemaphore DatManager::m_semap( 1 );

/* private */
/* This function searches instance of DatInfo specified by url.
   If instance does not exist, create instance. */
DatInfo* DatManager::getDatInfo( const KURL& url )
{

    m_semap++;

    if ( url.isEmpty() ) {
        m_semap--;
        return NULL;
    }

    int i = 0;
    DatInfoList::Iterator it;

    /* search */
    if ( m_datInfo.count() ) {
        for ( it = m_datInfo.begin(); it != m_datInfo.end(); ++it, i++ ) {

            if ( url.host() == ( *it ) ->url().host() &&
                    url.path() == ( *it ) ->url().path() ) {

                /* LRU */
                if ( i ) {
                    m_datInfo.remove( it );
                    m_datInfo.prepend( ( *it ) );
                }

                m_semap--;
                return ( *it );
            }
        }
    }

    /* not found */

    /*create new DatInfo and insert it into list. */
    KURL daturl = url.protocol() + "://" + url.host() + url.path();

    DatInfo* datInfo;
    datInfo = new DatInfo( daturl );
    if ( datInfo->getMaxResNumber() == 0 ) { /* cache does not exist */
        delete( datInfo );

        m_semap--;
        return NULL;
    }

    m_datInfo.prepend( datInfo );

    /* delete the last item of list */
    if ( m_datInfo.count() > DMANAGER_MAXQUEUE ) {
        it = m_datInfo.fromLast();
        m_datInfo.remove( it );

        delete( *it );
    }

    m_semap--;
    return datInfo;
}


/* public */
void DatManager::deleteDat( const KURL& url )
{
    m_semap++;

    DatInfoList::Iterator it;

    for ( it = m_datInfo.begin(); it != m_datInfo.end(); ++it ) {
        if ( url.host() == ( *it ) ->url().host() &&
                url.path() == ( *it ) ->url().path() ) {
            m_datInfo.remove( it );
            delete( *it );

            m_semap--;
            return ;
        }
    }

    m_semap--;
}


/* public */
const QString& DatManager::getDat( const KURL& url, int num )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    return datInfo->getDat( num );
}

/* public */
const QString& DatManager::getSubject( const KURL& url )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    return datInfo->getSubject();
}


/* public */
int DatManager::getNumByID( const KURL& url, const QString& strid )
{

    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return 0;

    int num = 0;
    int maxNum = datInfo->getMaxResNumber();

    for ( int i = 1; i <= maxNum; i++ ) {
        if ( datInfo->getId( i ) == strid ) num ++;
    }

    return num;
}


/* public */
int DatManager::getMaxResNumber( const KURL& url )
{
    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return 0;

    return datInfo->getMaxResNumber();
}


/* public */
QString DatManager::getHtml( const KURL& url, int startnum, int endnum )
{

    DatInfo * datInfo = getDatInfo( url );
    if ( datInfo == NULL ) return QString::null;

    QString retstr = QString::null ;

    for(int num = startnum; num <= endnum; num++)
	retstr += datInfo->getHtml(num);

    return retstr;
}


/* public */ /* This function is also called by kitadomtree. */
bool DatManager::checkAbone( const QString& str, int mode )
{

    QStringList & strlist = KitaConfig::aboneIDList();
    switch ( mode ) {
    case ABONECHK_BY_ID: break;
    case ABONECHK_BY_NAME: strlist = KitaConfig::aboneNameList(); break;
    case ABONECHK_BY_WORD: strlist = KitaConfig::aboneWordList(); break;
    default: return FALSE;
    }

    int i = -1;
    for ( QStringList::iterator it = strlist.begin();
            it != strlist.end(); ++it ) {
        i = str.find( ( *it ) );
        if ( i != -1 ) {
            return TRUE;
        }
    }

    return FALSE;
}


/* public */
QString DatManager::getHtmlByID(const KURL& url, const QString& strid, int &num ){

    DatInfo* datInfo = getDatInfo(url);
    if(datInfo == NULL) return QString::null;

    QString retstr = QString::null ;    
    num = 0;
    int maxNum = datInfo->getMaxResNumber();

    for ( int i = 1; i <= maxNum; i++ ) {
	if(datInfo->getId(i) == strid){
	    num ++;
	    retstr += datInfo->getHtml(i);
	}
    }

    return retstr;
}


/* public */
bool DatManager::checkID(const KURL& url, const QString& strid, int num ){

    DatInfo* datInfo = getDatInfo(url);
    if(datInfo == NULL) return FALSE;
    
    if(datInfo->getId(num) == strid) return TRUE;

    return FALSE;
}


/* check keywords */ /* public */
bool DatManager::checkWord(const KURL& url,
			    QStringList& stlist, int num,
			    bool checkOR /* AND or OR search */
    )
{
    DatInfo* datInfo = getDatInfo(url);
    if(datInfo == NULL) return FALSE;

    QString str_text = datInfo->getDat(num);
    if(str_text == QString::null) return FALSE;
    
    for ( QStringList::iterator it = stlist.begin(); it != stlist.end(); ++it ) {

        if ( checkOR ) { /* OR */
            if ( str_text.find( ( *it ), 0, FALSE ) != -1 ) {
                return TRUE;
            }
        } else { /* AND */
            if ( str_text.find( ( *it ), 0, FALSE ) == -1 ) return FALSE;
        }
    }

    if ( checkOR ) return FALSE;

    return TRUE;
}


/*------------------------------------------*/
/* parsing function for anchor (>>number)   */
/*------------------------------------------*/ /* public */

/* This fuction parses res-anchor.                          

   For example, if cdat = "&gt;12-20", then

   linkstr = ">12-20",
   refNum[0] = 12,
   refNum[1] = 20,
   pos = 9.

   This function is also called in KitaDomTree::createResAnchor. */

bool DatManager::parseResAnchor(

    /* input */
    const QChar *cdat, const unsigned int length,

    /* output */
    QString& linkstr, int* refNum, unsigned int& pos ){

    struct LocalFunc {
        static bool isHYPHEN( unsigned short c )
        {

            /* UTF-16 */
            if ( c == '-'
                    || ( c >= 0x2010 && c <= 0x2015 )
                    || ( c == 0x2212 )
                    || ( c == 0xFF0D )      /* UTF8: 0xEFBC8D */
               ) {
                return TRUE;
            }

            return FALSE;
        }
    };

    bool ret = FALSE;
    int i;
    
    linkstr = QString::null;
    refNum[0] = 0;
    refNum[1] = 0;    
    pos = 0;

    /* check '>' twice */
    for ( i = 0;i < 2;i++ ) {
	
        if ( cdat[ pos ].unicode() == UTF16_BRACKET ) {
            linkstr += cdat[ pos ];
            pos++;
        } else if ( cdat[ pos ] == '&' && cdat[ pos + 1 ] == 'g'  /* &gt; */
                    && cdat[ pos + 2 ] == 't' && cdat[ pos + 3 ] == ';' ) {
            linkstr += ">";
            pos += 4;
        }

    }

    /* check ',' */
    if ( !pos ) {
        if ( cdat[ pos ] == ',' || cdat[ pos ].unicode() == UTF16_COMMA ) {
            linkstr += ",";
            pos ++;
        }
    }

    /* check '=' */
    if ( !pos ) {
        if ( cdat[ pos ] == '=' || cdat[ pos ].unicode() == UTF16_EQ ) {
            linkstr += "=";
            pos ++;
        }
    }

    /* check numbers */
    int hyphen = 0;

    if( pos ){
    
	for ( i = 0 ; i < KITA_RESDIGIT + 1 && pos < length ; i++, pos++, ret = TRUE ) {

	    unsigned short c = cdat[ pos ].unicode();

	    if ( ( c < UTF16_0 || c > UTF16_9 )
		 && ( c < '0' || c > '9' )
		 && ( !LocalFunc::isHYPHEN( c )
		      || ( i == 0 && LocalFunc::isHYPHEN( c ) )
		      || ( hyphen && LocalFunc::isHYPHEN( c ) ) )
		) break;

	    linkstr += cdat[ pos ];

	    if ( LocalFunc::isHYPHEN( c ) ) {
		hyphen = 1;
		i = -1;
	    } else {
		if ( c >= UTF16_0 ) c = '0' + cdat[ pos ].unicode() - UTF16_0;
		refNum[hyphen] *= 10;
		refNum[hyphen] += c - '0';
	    }
	}
    }

    return ret;
}

/* public */
bool DatManager::checkRes(const KURL& url,const int target,const int num )
{
    DatInfo* datInfo = getDatInfo(url);
    if(datInfo == NULL) return FALSE;

    return datInfo->checkRes(target,num);
}


/* public */
QString DatManager::getTreeByRes(const KURL& url, const int resNum, int &num ){

    DatInfo* datInfo = getDatInfo(url);
    if(datInfo == NULL) return QString::null;

    QString retstr = QString::null ;    
    num = 0;

    QString tmp = QString().setNum( resNum );
    retstr = "<a href=\"#" + tmp + "\">&gt;&gt;" + tmp + "</a><br>";
    retstr += datInfo->getTreeByRes(resNum,num,"");

    return retstr;
}

